/**
 * SmartCGMS - continuous glucose monitoring and controlling framework
 * https://diabetes.zcu.cz/
 *
 * Copyright (c) since 2018 University of West Bohemia.
 *
 * Contact:
 * diabetes@mail.kiv.zcu.cz
 * Medical Informatics, Department of Computer Science and Engineering
 * Faculty of Applied Sciences, University of West Bohemia
 * Univerzitni 8, 301 00 Pilsen
 * Czech Republic
 *
 *
 * Purpose of this software:
 * This software is intended to demonstrate work of the diabetes.zcu.cz research
 * group to other scientists, to complement our published papers. It is strictly
 * prohibited to use this software for diagnosis or treatment of any medical condition,
 * without obtaining all required approvals from respective regulatory bodies.
 *
 * Especially, a diabetic patient is warned that unauthorized use of this software
 * may result into severe injure, including death.
 *
 *
 * Licensing terms:
 * Unless required by applicable law or agreed to in writing, software
 * distributed under these license terms is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * a) For non-profit, academic research, this software is available under the
 *      GPLv3 license.
 * b) For any other use, especially commercial use, you must contact us and
 *       obtain specific terms and conditions for the use of the software.
 * c) When publishing work with results obtained using this software, you agree to cite the following paper:
 *       Tomas Koutny and Martin Ubl, "Parallel software architecture for the next generation of glucose
 *       monitoring", Procedia Computer Science, Volume 141C, pp. 279-286, 2018
 */

#include "daily_view.h"

#include <iomanip>
#include <unordered_set>
#include <cmath>

#include <scgms/iface/UIIface.h>
#include <scgms/rtl/UILib.h>
#include <scgms/utils/drawing/IRenderer.h>
#include <scgms/utils/drawing/SVGRenderer.h>
#include <scgms/utils/string_utils.h>

#undef min
#undef max

NDrawing_Error CDaily_View::Draw(std::string& target, const TDraw_Options_Local& opts, const IDrawing_Data_Source& source)
{
	////// NOTE OP-E: this is just an example file to demonstrate the capabilities of SmartCGMS rendering SDK

	drawing::Drawing draw;

	auto& grp = draw.Root().Add<drawing::Group>("root");

	// axes
	////// NOTE OP-E: you would probably need to scale this with width/height, similarly to the graph_view.cpp routines
	grp.Add<drawing::Line>(50, opts.height, opts.width, opts.height)
		.Set_Stroke_Width(2);
	grp.Add<drawing::Line>(50, opts.height, 50, 50)
		.Set_Stroke_Width(2);


	// let's just draw a simple circle at 300;200, with a stroke and fill color and given opacities
	grp.Add<drawing::Circle>(300, 200, 8)
		.Set_Stroke_Color(RGBColor::From_UInt32(0xFFFF00FF, true))
		.Set_Fill_Color(RGBColor::From_UInt32(0xFFFFFF00, true))
		.Set_Stroke_Width(2)
		.Set_Stroke_Opacity(1.0)
		.Set_Fill_Opacity(1.0);



	// and another example: draw a polyline
	drawing::PolyLine line;

	line.Add_Point(100, 100);
	line.Add_Point(150, 120);
	line.Add_Point(250, 100);
	line.Add_Point(350, 200);
	line.Add_Point(100, 190);

	const auto& all_points = line.Get_Points();
	if (!all_points.empty())
		line.Set_Position(all_points[0].x, all_points[0].y);

	grp.Add<drawing::PolyLine>(line)
		.Set_Stroke_Width(1.2)
		.Set_Fill_Opacity(0)
		.Set_Stroke_Color(RGBColor::From_UInt32(0xFF00FFFF, true));






	////// NOTE OP-E: this is a basic loop over all segments - this part should probably remain as-is, you would modify just the parts below
	for (const auto seg_id : opts.segment_ids) {

		const TPlot_Segment& segment = source.Get_Segment(seg_id);

		for (const auto& signal : segment.mPlots_Signals) {

			if (opts.signal_ids.find(signal.first) == opts.signal_ids.end())
				continue;

			scgms::TSignal_Descriptor desc = scgms::Null_Signal_Descriptor;

			if (scgms::get_signal_descriptor_by_id(signal.second.signal_id, desc))
			{
				const auto values = signal.second.mPlots_Values;

				////// NOTE OP-E: description follows

				// everything you need to know about the signal is now in the "desc" instance
				// e.g., desc.signal_description contains the name of the signal, desc.visualization contains the type of visualization (smooth curve, mark, ...), etc.
				// colors are coded in standard RGBA format (as integers), to convert them to SVG representation, use RGBColor::From_UInt32(...)
				// times are coded in "rat-time", where the whole part represent days, and the fractional part represents time of the day, so e.g., 1.5 = 1 day and 12 hours

				// the signal values are stored in "values" vector, you can iterate through it as follows:
				for (auto val : values) {
					// Now the "time" is in val.device_time
					// The value is in val.value
				}
			}

		}
	}

	// render prepared canvas to string
	CSVG_Renderer renderer(opts.width, opts.height, target);
	draw.Render(renderer);

	return NDrawing_Error::Ok;
}
